'''Sets properties of atoms'''

import pandas as pd
from .readWriteFiles import getElementData

def setAtomicRadius(atomRadius):
    '''Defines the dataframe to find the atomic radius'''
    radii = {}

    ele = getElementData(atomRadius)

    radii = pd.Series(ele[atomRadius].values, index=ele.Symbol).to_dict()

    # Convert radii to Angstroms
    for k, v in radii.items():
        if v != 'nan':
            radii[k] = v * 0.01

    return radii

def setAtomicMass():
    '''Sets the mass of an atom'''
    mass = {}
    ele = getElementData('AtomicMass')
    mass = pd.Series(ele.AtomicMass.values, index=ele.Symbol).to_dict()

    return mass



'''Module places molecules in a sphere'''

import random
import scipy.spatial
import numpy as np
from .isOverlap import isOverlapAtomKDTree, isOverlapMoleculeKDTree, buildKDTreeMapping
from .makeStruc import makeBase, reCenter, randReorient

def atomFillSphere(numShifts, sphere, og, radii, tol, numMol):
    '''Fills sphere with single atoms'''
    filled = []

    if str(numMol).lower() == 'fill':
        numMol = 10000000000000

    # Create KD-tree for filledAtoms
    kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

    for zShifts in range(numShifts):
        for yShifts in range(numShifts):
            for xShifts in range(numShifts):
                for atom in og:
                    # Calculate the relative position of each atom within the molecule
                    relX = atom[1]
                    relY = atom[2]
                    relZ = atom[3]
                    atomType = atom[0]

                    # Calculate the coordiantes within the sphere
                    x = (sphere.xCoord - sphere.radius) + relX + xShifts
                    y = (sphere.yCoord - sphere.radius) + relY + yShifts
                    z = (sphere.zCoord - sphere.radius) + relZ + zShifts

                    # Adjust for atomic radii
                    atomRadius = radii.get(atomType, 0.0) # Get the radius for the atom type

                    # Check if the new atom fits within the sphere
                    if sphere.containsPoints(x, y, z, atomRadius):
                        if len(atom) == 5:
                            newAtom = (atom[0], x, y, z, atom[4])
                        else:
                            newAtom = (atom[0], x, y, z)

                        if (kdTree is None or not isOverlapAtomKDTree(newAtom, kdTree, indexToAtom, radii, tol)) and \
                            numMol > len(filled):
                            filled.append(newAtom)

                            # Rebuild KDTree with newly added atoms
                            kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

    return filled

def atomRandSphere(numMol, maxAttempts, og, sphere, radii, tol):
    '''Places a defined number of atoms in a sphere'''
    filled = []
    attempts = 0

    # Create KD-tree for filledAtoms
    kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

    while len(filled) < numMol and attempts <= maxAttempts:
        for atom in og:

            # Calculate the shift for each tile and new point
            newX = atom[1] + random.uniform((sphere.xCoord - sphere.radius),
                                            (sphere.xCoord + sphere.radius))
            newY = atom[2] + random.uniform((sphere.yCoord - sphere.radius),
                                            (sphere.yCoord + sphere.radius))
            newZ = atom[3] + random.uniform((sphere.zCoord - sphere.radius),
                                                (sphere.zCoord + sphere.radius))
            atomType = atom[0]

            # Adjust for atomic radii
            atomRadius = radii.get(atomType, 0.0) # Get the radius for the atom type

            # Check if the new atom fits within the sphere
            if sphere.containsPoints(newX, newY, newZ, atomRadius):
                if len(atom) == 5:
                    newAtom = (atom[0], newX, newY, newZ, atom[4])
                else:
                    newAtom = (atom[0], newX, newY, newZ)

                if (kdTree is None or not isOverlapAtomKDTree(newAtom, kdTree, indexToAtom, radii, tol)):
                    filled.append(newAtom)

                    # Rebuild KDTree with newly added atoms
                    kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

        attempts += 1

    return list(filled)

def moleculeFillSphere(numShifts, sphere, og, radii, tol,
                       baseStruc, randOrient, numMol):
    '''Fills sphere with molecules'''
    filled = []

    if str(numMol).lower() == 'fill':
        numMol = 10000000000000

    if baseStruc is not None:
        base = makeBase(baseStruc)
        filled.append(reCenter(base, sphere))

    # Create KD-tree for filledAtoms
    kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

    for zShifts in range(numShifts):
        for yShifts in range(numShifts):
            for xShifts in range(numShifts):
                newMol = []

                for atom in og:
                    # Calculate the relative position of each atom within the molecule
                    relX = atom[1]
                    relY = atom[2]
                    relZ = atom[3]
                    atomType = atom[0]

                    # Calculate the coordiantes within the sphere
                    x = (sphere.xCoord - sphere.radius) + relX + xShifts
                    y = (sphere.yCoord - sphere.radius) + relY + yShifts
                    z = (sphere.zCoord - sphere.radius) + relZ + zShifts

                    # Adjust for atomic radii
                    atomRadius = radii.get(atomType, 0.0) # Get the radius for the atom type

                    # Check if the new atom fits within the sphere
                    if sphere.containsPoints(x, y, z, atomRadius):
                        if len(atom) == 5:
                            newAtom = (atom[0], x, y, z, atom[4])
                        else:
                            newAtom = (atom[0], x, y, z)
                        newMol.append(newAtom)
                    else:
                        break # If any atom doesn't fit, discard the whol molecule

                if randOrient and len(newMol) == len(og):
                    newMol = randReorient(newMol)

                if (kdTree is None or not isOverlapMoleculeKDTree(newMol, kdTree, indexToAtom, radii, tol)) and \
                    numMol > len(filled):
                    if len(newMol) == len(og):
                        filled.append(newMol)

                        # Rebuild KDTree with newly added atoms
                        kdTree, indexToAtom = buildKDTreeMapping(filled, radii)
    return filled

def moleculeRandSphere(numMol, maxAttempts, og, sphere, radii, tol,
                           baseStruc, randOrient):
    '''Places a defined number of molecules in a sphere'''
    filled = []
    attempts = 0

    if baseStruc is not None:
        base = makeBase(baseStruc)
        filled.append(reCenter(base, sphere))

    # Create KD-tree for filledAtoms
    kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

    while len(filled) < numMol and attempts <= maxAttempts:
        newMol = []

        shiftX = 0
        shiftY = 0
        shiftZ = 0

        shiftX = random.uniform((sphere.xCoord - sphere.radius),
                                (sphere.xCoord + sphere.radius))
        shiftY = random.uniform((sphere.yCoord - sphere.radius),
                                (sphere.yCoord + sphere.radius))
        shiftZ = random.uniform((sphere.zCoord - sphere.radius),
                                    (sphere.zCoord + sphere.radius))
        for atom in og:

            # Calculate the shift for each tile and new point
            newX = atom[1] + shiftX
            newY = atom[2] + shiftY
            newZ = atom[3] + shiftZ
            atomType = atom[0]

            # Adjust for atomic radii
            atomRadius = radii.get(atomType, 0.0) # Get the radius for the atom type
            # Check if the new atom fits within the sphere
            if sphere.containsPoints(newX, newY, newZ, atomRadius):
                if len(atom) == 5:
                    newAtom = (atom[0], newX, newY, newX, atom[4])
                else:
                    newAtom = (atom[0], newX, newY, newZ)
                newMol.append(newAtom)
            else:
                break # If any atom doesn't fit, discard the whol molecule

        if randOrient and len(newMol) == len(og):
            newMol = randReorient(newMol)

        if (kdTree is None or not isOverlapMoleculeKDTree(newMol, kdTree, indexToAtom, radii, tol)):
            if len(newMol) == len(og):
                filled.append(newMol)

                # Rebuild KDTree with newly added atoms
                kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

        attempts += 1

    return list(filled)



'''
This module is for reading and writing files as needed. 
Since both the input and output files are optional, these functions can be avoided.
The structure file is required however, and currently the only supported format is xyz files.
Data on elements is also imported and searchable. Primary use at this time is for the atomic radius
'''

import os
import pandas as pd
import trimesh

def getInput(filePath):
    '''Reads input file if provided'''
    params = {}
    with open(filePath, encoding = 'utf-8') as f:
        for line in f:
            key, value = line.strip().split('=')
            params[key.strip()] = value.strip()

    return params

def readStrucFile(filePath):
    '''Reads structure file'''
    # Reads XYZ
    if filePath[-3:] == 'xyz':
        return pd.read_csv(filePath, delim_whitespace=True,
                           skiprows=2, names=["Atom", "X", "Y", "Z"]), None
    # Reads PDB
    if filePath[-3:] == 'pdb':
        return readPdb(filePath)

    return f"ERROR: Issue generating file to {filePath}"

def readPdb(filePath):
    '''Reads structure file if a pdb'''
    data = []
    unitCell = None
    symbols = getElementData('AtomicMass')

    with open(filePath, 'r', encoding = 'utf-8') as pdbFile:
        for line in pdbFile:

            if line.startswith('CRYST1'):
                # Extract unit cell parameters from the CRYST1 line
                a = float(line[6:15].strip())
                b = float(line[15:24].strip())
                c = float(line[24:33].strip())
                alpha = float(line[33:40].strip())
                beta = float(line[40:47].strip())
                gamma = float(line[47:54].strip())
                unitCell = {'a': a, 'b': b, 'c': c, 'alpha': alpha, 'beta': beta, 'gamma': gamma}

            elif line.startswith('ATOM') or line.startswith('HETATM'):
                # Parse relevant fields from the PDB format
                atomName = line[12:16].strip()
                element = line[76:78].strip()

                if element not in symbols['Symbol']:
                    element = atomName

                try:
                    x = float(line[30:38])
                    y = float(line[38:46])
                    z = float(line[46:54])
                except ValueError:
                    continue # Skip line if coordinates are invalid
                residueName = line[17:20].strip()
                residueSeq = int(line[22:26].strip())

                data.append([element, x, y, z, residueName, residueSeq])

    df = pd.DataFrame(data, columns=['Atom', 'X', 'Y', 'Z', 'Residue', 'ResidueSeq'])

    return df, unitCell

def readMesh(filePath):
    '''Reads mesh files in format like .obj, .stl, .ply'''
    mesh = trimesh.load(filePath)
    return mesh

def writeOutput(data, filePath, strucType, cellParams=None):
    '''Writes data to output file'''
    try:
        # Writes XYZ
        if filePath[-3:] == 'xyz':
            writeXYZ(data, filePath, strucType)
        # Writes PDB
        elif filePath[-3:] == 'pdb':
            if cellParams != None:
                writePdb(data, filePath, cellParams)
            else:
                writePdb(data, filePath)
    except KeyError:
        print(f"Filetype {filePath[-3:]} not supported\n")

def writePdb(data, filePath, cellParams=None):
    '''Writes a pdb file from results'''
    template = (
        "HETATM{atomNum:5d} {atomType:>2}  {residueName:>3} A{resNum: >4d}"
        "    {x: >8.3f}{y: >8.3f}{z: >8.3f}{occupancy: >6.2f}{tempFactor: >6.2f}\n"
        )

    with open(filePath, 'w', encoding = 'utf-8') as pdbFile:
        atomNum = 1
        resNum = 1
        cell = cellParams

        if cell != None:
            pdbFile.write(cell + '\n')

        for mol in data:
            for atom in mol:
                pdbFile.write(template.format(
                   atomNum = atomNum,
                   atomType = atom[0],
                   residueName = atom[4],
                   resNum = resNum,
                   x = atom[1],
                   y = atom[2],
                   z = atom[3],
                   occupancy = 1.00, # Default value
                   tempFactor = 0.00, # Default value
                ))

                atomNum += 1
            resNum += 1
        pdbFile.write("END\n")

def writeXYZ(data, filePath, strucType):
    '''Writes an xyz file from results'''
    if strucType == 'molecule':
        columns = ['Atom', 'X', 'Y', 'Z']
        df = pd.DataFrame(columns=['Atom', 'X', 'Y', 'Z'])

        for mol in data:
            for atom in mol:
                df = df._append(pd.Series(atom, index=columns),
                                ignore_index = True)

        with open(filePath, 'w', encoding = 'utf-8') as f:
            f.write(str(len(df['Atom'])))
            f.write('\n\n')
            f.write(df.to_string(header=False, index=False))

    else:
        df = pd.DataFrame(data, columns=['Atom', 'X', 'Y', 'X'])
        with open(filePath, 'w', encoding = 'utf-8') as f:
            f.write(str(len(df['Atom'])))
            f.write('\n\n')
            f.write(df.to_string(header=False, index=False))

def getElementData(param):
    '''
    Import Atomic Data
    Database Information
    Mass: g/mol
    Denisty Units: g/mL, Temp Units: Kelvin, Heat Capacity Units: J/(g*K),
    Heat of Vaporization/Fusion Units = kJ/mol
    Molar Volume Units: mL, Thermal Conductivty Units: Watts/(m * K),
    Atomic/Covalent/Van Det Waals Radius Units = pm
    Ionization Energy Units = kJ/mol, Electron Affinity Units = kJ/mol
    Data From: https://github.com/Bluegrams/periodic-table-data
    '''

    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, 'data/ElementData.csv')

    elements = pd.read_csv(filename, delimiter=",", header=0)
    columnsToRemove = ["DiscoveredBy", "Discovery", "AbundanceCrust", "AbundanceUniverse"]
    elements = elements.drop(labels=columnsToRemove, axis = 1)

    return pd.DataFrame(elements, columns=['Symbol', param])



'''
This module handles a 3d box/cube to be filled
'''

class Box3d:
    '''
    Defines the 3D Box object
    '''
    def __init__(self, xCoord, yCoord, zCoord, dims):
        self.xCoord = xCoord
        self.yCoord = yCoord
        self.zCoord = zCoord
        self.length = dims[0]
        self.width = dims[1]
        self.height = dims[2]

    def volume(self):
        '''Returns the volume of a box'''
        return self.length * self.width * self.height

    def surfaceArea(self):
        '''Returns the surface area of a box'''
        face1 = 2 * self.length * self.width
        face2 = 2 * self.length * self.height
        face3 = 2 * self.width * self.height

        return face1 + face2 + face3

    def findCenter(self):
        '''Finds the center point of the Box'''
        corner1 = [self.xCoord, self.yCoord, self.zCoord]
        corner2 = [self.length, self.width, self.height]

        centerX = (corner1[0] + corner2[0]) / 2
        centerY = (corner1[1] + corner2[1]) / 2
        centerZ = (corner1[2] + corner2[2]) / 2

        return (centerX, centerY, centerZ)

    def origin(self):
        '''Find the origin'''
        return (self.xCoord, self.yCoord, self.zCoord)

def drawBox(params):
    '''
    Function sets the array that describes the dimensions of the box
    based on if the box is a cube or not
    '''
    if params['shape'].lower() == 'cube':
        return [float(params['sideLength']), float(params['sideLength']),
                float(params['sideLength'])]

    if params['shape'].lower() == 'box':
        return [float(params['Xlen']), float(params['Ylen']), float(params['Zlen'])]

    return 'ERROR: Shape not found\n'



'''Places molecules in a given volume based on the provided denisty'''

import numpy as np
import scipy.spatial
from .makeStruc import calcDistance, randReorient
from .isOverlap import isOverlapMoleculeKDTree, isOverlapAtomKDTree, buildKDTreeMapping

def placeMols(shape, og, density, tol, shapeType, radii, randOrient):
    '''Place molecules in the grid defined by the density'''
    gridPoints = calcDistance(shape, og, density, shapeType)

    def calcCentroid(coords):
        """Calculate the geometric centroid of a set of coordinates."""
        return np.mean(coords, axis=0)

    # Tanslate molecules such that their centroid is at the center
    # of the target point
    def translateMol(coords, target):
        '''Translate molecules to a new grid point'''
        atomLabels = [atom[0] for atom in coords]
        pdbInfo = [atom[4] if len(atom) == 5 else None for atom in coords]
        xyzCoords = np.array([atom[1:4] for atom in coords])

        centroid = calcCentroid(xyzCoords)
        transVec = np.array(target) - centroid
        translatedXYZ = xyzCoords + transVec

        transMol = [[label] + coord.tolist() + ([info] if info is not None else [])
                    for label, coord, info in zip(atomLabels, translatedXYZ, pdbInfo)]
        
        return transMol

    strucType = 'molecule'
    mols = []

    for point in gridPoints:
        newPoint = translateMol(og, point)

        if len(mols) == 0:
            mols.append(newPoint)

            # Create KD-tree for filledAtoms
            kdTree, indexToAtom = buildKDTreeMapping(mols, radii)

        if randOrient:
            newPoint = randReorient(newPoint)

        if len(og) == 1:
            if (kdTree is None or not isOverlapAtomKDTree(newPoint, kdTree, indexToAtom, radii, tol)):
                mols.append(newPoint)

                # Rebuild KDTree with newly added atoms
                kdTree, indexToAtom = buildKDTreeMapping(mols, radii)

                strucType = "atom"
        else:
            if (kdTree is None or not isOverlapMoleculeKDTree(newPoint, kdTree, indexToAtom, radii, tol)):
                mols.append(newPoint)

                # Rebuild KDTree with newly added atoms
                kdTree, indexToAtom = buildKDTreeMapping(mols, radii)

                strucType = "molecule"
    
    return mols, strucType



# Top level package
from .Box3D import Box3d, drawBox
from .defaultParams import defaultParams
from .drawMol import drawMolBox, drawMolSphere, drawMolMesh
from .isOverlap import isOverlapAtom, isOverlapMolecule, isOverlapAtomKDTree, isOverlapMoleculeKDTree, buildKDTreeMapping
from .makeStruc import makeBase, calcCenter, reCenter, shiftPoints, randReorient, calcDensity, calcNumMol, calcDistance
from .readWriteFiles import getInput, readStrucFile, readPdb, writeOutput, writePdb, writeXYZ, getElementData, readMesh
from .setAtomProp import setAtomicMass, setAtomicRadius
from .shiftBox import atomsFillBox, atomsRandBox, moleculesFillBox, moleculesRandBox, inBox
from .shiftSphere import atomFillSphere, atomRandSphere, moleculeFillSphere, moleculeRandSphere
from .Sphere3D import Sphere3d
from .shiftDensity import placeMols
from .mesh3D import mesh3D
from .shiftMesh import atomsFillMesh, moleculesFillMesh, atomsRandMesh, moleculesRandMesh
from .shiftUnitCell import unitCellBox, unitCellSphere, unitCellMesh

from . import _version
__version__ = _version.get_versions()['version']



'''Module places molecules in a box'''

import random
import scipy.spatial
import numpy as np
from .isOverlap import isOverlapAtomKDTree, isOverlapMoleculeKDTree, buildKDTreeMapping
from .makeStruc import makeBase, reCenter, randReorient

def atomsFillBox(x, y, z, tol, og, box, radii, numMol):
    '''Fills box with single atoms'''
    if str(numMol).lower() == 'fill':
        numMol = 10000000000000

    filledAtom = []

    # Create KD-tree for filledAtoms
    kdTree, indexToAtom = buildKDTreeMapping(filledAtom, radii)

    for zShift in range(z):
        for yShift in range(y):
            for xShift in range(x):
                for atom in og:
                    # Calculate the shift for each tile and new point
                    newX = box.xCoord + atom[1] + xShift
                    newY = box.yCoord + atom[2] + yShift
                    newZ = box.zCoord + atom[3] + zShift

                    # Adjust for atomic radiss
                    newXMin = newX - radii[atom[0]]
                    newYMin = newY - radii[atom[0]]
                    newZMin = newZ - radii[atom[0]]
                    newXMax = newX + radii[atom[0]]
                    newYMax = newY + radii[atom[0]]
                    newZMax = newZ + radii[atom[0]]

                    # Check if the new atom fits within the box
                    if inBox(newXMin, newXMax, newYMin, newYMax, newZMin, newZMax, box):
                        if len(atom) == 5:
                            newAtom = (atom[0], newX, newY, newZ, atom[4])
                        else:
                            newAtom = (atom[0], newX, newY, newZ)

                        if (kdTree is None or not isOverlapAtomKDTree(newAtom, kdTree, indexToAtom, radii, tol)) and \
                            numMol > len(filledAtom):
                            filledAtom.append(newAtom)

                            # Rebuild KDTree with newly added atoms
                            kdTree, indexToAtom = buildKDTreeMapping(filledAtom, radii)

    return filledAtom

def atomsRandBox(numMol, maxAttempts, og, box,
                     radii, tol):
    '''Places a defined number of single atoms in box'''
    filledAtom = []

    # Create KD-tree for filledAtoms
    kdTree, indexToAtom = buildKDTreeMapping(filledAtom, radii)

    attempts = 0

    while len(filledAtom) < numMol and attempts < maxAttempts:
        for atom in og:
            # Calculate the shift for each tile and new point
            newX = atom[1] + random.uniform(0, box.length)
            newY = atom[2] + random.uniform(0, box.width)
            newZ = atom[3] + random.uniform(0, box.height)
            # Adjust for atomic radiss
            newXMin = newX - radii[atom[0]]
            newYMin = newY - radii[atom[0]]
            newZMin = newZ - radii[atom[0]]
            newXMax = newX + radii[atom[0]]
            newYMax = newY + radii[atom[0]]
            newZMax = newZ + radii[atom[0]]
            # Check if the new atom fits within the box
            if  inBox(newXMin, newXMax, newYMin, newYMax, newZMin, newZMax, box):
                if len(atom) == 5:
                    newAtom = (atom[0], newX, newY, newZ, atom[4])
                else:
                    newAtom = (atom[0], newX, newY, newZ)

                if (kdTree is None or not isOverlapAtomKDTree(newAtom, kdTree, indexToAtom, radii, tol)):
                    filledAtom.append(newAtom)

                    # Rebuild KDTree with newly added atoms
                    kdTree, indexToAtom = buildKDTreeMapping(filledAtom, radii)

        attempts += 1

    return list(filledAtom)

def moleculesFillBox(x, y, z, tol, og, box, radii,
                     baseStruc, randOrient, numMol):
    '''Fills box with molecules'''
    filledAtom = []

    if str(numMol).lower() == 'fill':
        numMol = 10000000000000

    if baseStruc is not None:
        base = makeBase(baseStruc)
        filledAtom.append(reCenter(base, box))

    kdTree, indexToAtom = buildKDTreeMapping(filledAtom, radii)

    for zShift in range(z):
        for yShift in range(y):
            for xShift in range(x):
                newMol = []
                anyAtomOutside = False

                for atom in og:
                    # Calculate the shift for each tile and new point
                    newX = box.xCoord + atom[1] + xShift
                    newY = box.yCoord + atom[2] + yShift
                    newZ = box.zCoord + atom[3] + zShift

                    # Adjust for atomic radiss
                    newXMin = newX - radii[atom[0]]
                    newYMin = newY - radii[atom[0]]
                    newZMin = newZ - radii[atom[0]]
                    newXMax = newX + radii[atom[0]]
                    newYMax = newY + radii[atom[0]]
                    newZMax = newZ + radii[atom[0]]

                    # Check if the new atom fits within the box
                    if  inBox(newXMin, newXMax, newYMin, newYMax, newZMin,
                              newZMax, box):
                        if len(atom) == 5:
                            newAtom = (atom[0], newX, newY, newZ, atom[4])
                        else:
                            newAtom = (atom[0], newX, newY, newZ)
                        newMol.append(newAtom)

                    else:
                        anyAtomOutside = True

                if randOrient and len(newMol) == len(og):
                    newMol = randReorient(newMol)

                if ((kdTree is None or not isOverlapMoleculeKDTree(newMol, kdTree, indexToAtom, radii, tol)) and
                    not anyAtomOutside and numMol > len(filledAtom)):

                    filledAtom.append(newMol)

                    # Rebuild KDTree with newly added atoms
                    kdTree, indexToAtom = buildKDTreeMapping(filledAtom, radii)

    return filledAtom

def moleculesRandBox(numMol, maxAttempts, og, box, radii, tol,
                         baseStruc, randOrient):
    '''Places a defined number of molecules in box'''
    filledAtom = []

    attempts = 0

    if baseStruc is not None:
        base = makeBase(baseStruc)
        filledAtom.append(reCenter(base, box))

    # Create KD-tree for filledAtoms
    kdTree, indexToAtom = buildKDTreeMapping(filledAtom, radii)

    while len(filledAtom) < numMol and attempts < maxAttempts:
        newMol = []
        anyAtomOutside = False
        xShift = random.uniform(0, box.length)
        yShift = random.uniform(0, box.width)
        zShift = random.uniform(0, box.height)
        for atom in og:
            # Calculate the shift for each tile and new point
            newX = atom[1] + xShift
            newY = atom[2] + yShift
            newZ = atom[3] + zShift
            # Adjust for atomic radiss
            newXMin = newX - radii[atom[0]]
            newYMin = newY - radii[atom[0]]
            newZMin = newZ - radii[atom[0]]
            newXMax = newX + radii[atom[0]]
            newYMax = newY + radii[atom[0]]
            newZMax = newZ + radii[atom[0]]
            # Check if the new atom fits within the box
            if  inBox(newXMin, newXMax, newYMin, newYMax, newZMin, newZMax, box):
                if len(atom) == 5:
                    newAtom = (atom[0], newX, newY, newZ, atom[4])
                else:
                    newAtom = (atom[0], newX, newY, newZ)
                newMol.append(newAtom)
            else:
                anyAtomOutside = True

        if randOrient and len(newMol) == len(og):
            newMol = randReorient(newMol)

        if (kdTree is None or not isOverlapMoleculeKDTree(newMol, kdTree, indexToAtom, radii, tol)) and not anyAtomOutside:
            filledAtom.append(newMol)

            # Rebuild KDTree with newly added atoms
            kdTree, indexToAtom = buildKDTreeMapping(filledAtom, radii)
        attempts += 1

    return list(filledAtom)

def inBox(xMin, xMax, yMin, yMax, zMin, zMax, box):
    '''Checks if atoms are fully in the box'''
    if (
        xMin >= box.xCoord and xMax <= box.xCoord + box.length and
        yMin >= box.yCoord and yMax <= box.yCoord + box.width and
        zMin >= box.zCoord and zMax <= box.zCoord + box.height
    ):
        return True
    return False



'''Defines mesh'''

import numpy as np
from .readWriteFiles import readMesh
import trimesh

class mesh3D():
    '''Defines properties of a mesh'''

    def __init__(self, filePath, scale):
        self.mesh = readMesh(filePath)
        self.scale = float(scale)
        self.mesh = self.mesh.apply_scale(self.scale)

        # Mesh properties
        self.isWaterTight = self.mesh.is_watertight
        self.bounds = self.mesh.bounds
        self.meshOrigin = self.origin()

        self.translate()

        # Scaled mesh
        self.meshBound = self.mesh.apply_scale((1.1, 1.1, 1))

    def isInside(self, point):
        '''Check if point is inside the mesh or on the boundary'''
        # Checks if the point is inside the mesh
        inside = self.meshBound.contains([point])[0]

        # Check if the point is on the boundary of the mesh
        # We will get the closest point on the mesh to our point and check if they are the same
        closestPoint, distance, _ = trimesh.proximity.closest_point(self.mesh, [point])
        onBound = np.isclose(distance[0], 0, atol=1e-3)

        return inside or onBound
    
    def volume(self):
        return self.mesh.volume
    
    def surfaceArea(self):
        return self.mesh.area
    
    def origin(self):
        return [0, 0, 0]
    
    def findCenter(self):
        return self.origin
    
    def translate(self):
        self.mesh.apply_translation(-self.bounds[0])
        self.bounds = self.mesh.bounds


'''Tests for package'''

# pylin: skip-file

from TurtleMol.drawMol import drawMolBox, drawMolSphere
from TurtleMol.readWriteFiles import readStrucFile
from TurtleMol.defaultParams import defaultParams

def testBaseStruc():
    iparamsSphere = {
        'shape' : 'sphere',
        'radius' : 5.0,
        'numMolecules' : 'fill',
        'structureFile' : "TurtleMol/examples/Methane_Water/water.xyz",
        'baseStrucFile' : "TurtleMol/examples/Methane_Water/ch4.xyz"
    }

    iparamsCube = {
        'shape' : 'cube',
        'sideLength' : 10.0,
        'numMolecules' : 'fill',
        'structureFile' : "TurtleMol/examples/Methane_Water/water.xyz",
        'baseStrucFile' : "TurtleMol/examples/Methane_Water/ch4.xyz"
    }

    dparams = defaultParams()

    for name in dparams:
        if name not in iparamsSphere:
            iparamsSphere[name] = dparams[name]

        if name not in iparamsCube:
            iparamsCube[name] = dparams[name]

    struc, unitCell = readStrucFile(iparamsSphere['structureFile'])

    if iparamsSphere['baseStrucFile']:
        baseStruc, baseUnitCell = readStrucFile(iparamsSphere['baseStrucFile'])
    else:
        baseStruc = None

    outStrucCube, strucTypeCube = drawMolBox(struc, baseStruc, iparamsCube)
    assert len(outStrucCube) != 0, "Output structure should have some length"
    assert strucTypeCube == "molecule", "Structure type should be an atom"

    outStrucSphere, strucTypeSphere = drawMolSphere(struc, baseStruc, iparamsSphere)
    assert len(outStrucSphere) != 0, "Output structure should have some length"
    assert strucTypeSphere == "molecule", "Structure type should be an atom"



'''This module is for replicating a unit cell if it is defined in a given pdb'''

import numpy as np
import trimesh
from .shiftBox import inBox

# Box
def unitCellBox(shape, dims, cellDims, og, radii):
    '''Duplicates unit cells to fill a given box'''

    # Find the atom types in the tile
    atomNames = {atom[0] for atom in og}
    totalRadius = sum(radii[str.capitalize(name)] for name in atomNames)

    # Calculate how many times to duplicate the unit cell in a given dimension
    dupeCount = [int(dims[i] / cellDims[i]) for i in range(3)]
    cellParams = f'CRYST1    {dupeCount[0]*cellDims[0]: .3f}    {dupeCount[1]*cellDims[1]: .3f}    {dupeCount[2]*cellDims[2]: .3f}  90.00  90.00  90.00 P1          1'

    filled = []

    for dx in range(dupeCount[0]):
        for dy in range(dupeCount[1]):
            for dz in range(dupeCount[2]):
                # Calculate the displacement for this duplication
                disp = np.array([dx * cellDims[0], 
                                 dy * cellDims[1], 
                                 dz * cellDims[2]])
                
                currentCell = []
                
                for atom in og:
                    newX = shape.xCoord + atom[1] + disp[0]
                    newY = shape.yCoord + atom[2] + disp[1]
                    newZ = shape.zCoord + atom[3] + disp[2]
                    atomType = atom[0]

                    atomRadius = radii.get(atomType, 0.0)

                    # Adjust for atomic radiss
                    newXMin = newX - atomRadius
                    newYMin = newY - atomRadius
                    newZMin = newZ - atomRadius
                    newXMax = newX + atomRadius
                    newYMax = newY + atomRadius
                    newZMax = newZ + atomRadius

                    # Check if the new atom fits within the box
                    if  inBox(newXMin, newXMax, newYMin, newYMax, newZMin,
                              newZMax, shape):
                        if len(atom) == 5:
                            newAtom = (atom[0], newX, newY, newZ, atom[4])
                        else:
                            newAtom = (atom[0], newX, newY, newZ)
                        currentCell.append(newAtom)

                filled.append(currentCell)
    return filled, "molecule", cellParams

def unitCellSphere(shape, cellDims, og, radii):
    '''Duplicates unit cells to fill a given sphere'''

    # Box dimensions that completely contain the sphere
    boxDim = [2 * shape.radius] * 3
    dupeCount = [int(boxDim[i] / cellDims[i]) for i in range(3)]
    cellParams = f'CRYST1    {dupeCount[0]*cellDims[0]: .3f}    {dupeCount[1]*cellDims[1]: .3f}    {dupeCount[2]*cellDims[2]: .3f}  90.00  90.00  90.00 P1          1'

    filled = []

    for dx in range(dupeCount[0]):
        for dy in range(dupeCount[1]):
            for dz in range(dupeCount[2]):
                disp = np.array([dx * cellDims[0], 
                                 dy * cellDims[1], 
                                 dz * cellDims[2]])
                currentCell = []

                for atom in og:
                    newX = (shape.xCoord - shape.radius)+ atom[1] + disp[0]
                    newY = (shape.yCoord - shape.radius) + atom[2] + disp[1]
                    newZ = (shape.zCoord - shape.radius) + atom[3] + disp[2]
                    atomType = atom[0]

                    atomRadius = radii.get(atomType, 0.0)

                    # Check if the new atom fits within the sphere
                    if shape.containsPoints(newX, newY, newZ, atomRadius):
                        if len(atom) == 5:
                            newAtom = (atom[0], newX, newY, newZ, atom[4])
                        else:
                            newAtom = (atom[0], newX, newY, newZ)
                        currentCell.append(newAtom)
                filled.append(currentCell)
    return filled, "molecule", cellParams

def unitCellMesh(shape, cellDims, og, radius):
    '''Duplicates unit cells to fill a given mesh'''

    # Find the atom types in the tile
    atomNames = {atom[0] for atom in og}
    totalRadius = sum(radius[str.capitalize(name)] for name in atomNames)

    # Box dimensions that completely contain the mesh
    maxBound, minBound = shape.bounds[1], shape.bounds[0]
    boxDim = maxBound - minBound
    dupeCount = [int(boxDim[i] / cellDims[i]) for i in range(3)]
    
    cellParams = f'CRYST1    {dupeCount[0]*cellDims[0]: .3f}    {dupeCount[1]*cellDims[1]: .3f}    {dupeCount[2]*cellDims[2]: .3f}  90.00  90.00  90.00 P1          1'

    filled = []

    for dx in range(dupeCount[0]):
        for dy in range(dupeCount[1]):
            for dz in range(dupeCount[2]):
                disp = np.array([dx * cellDims[0], 
                                 dy * cellDims[1], 
                                 dz * cellDims[2]]) + minBound
                currentCell = []

                for atom in og:
                    newX = shape.origin()[0] + atom[1] + disp[0]
                    newY = shape.origin()[1] + atom[2] + disp[1]
                    newZ = shape.origin()[2] + atom[3] + disp[2]
                    atomType = atom[0]

                    if shape.isInside([newX, newY, newZ]):
                        if len(atom) == 5:
                            newAtom = (atom[0], newX, newY, newZ, atom[4])
                        else:
                            newAtom = (atom[0], newX, newY, newZ)
                        currentCell.append(newAtom)
                filled.append(currentCell)
    return filled, "molecule", cellParams


'''Fills an arbitrary mesh with molecules'''

import random
import numpy as np
from .isOverlap import isOverlapAtomKDTree, isOverlapMoleculeKDTree, buildKDTreeMapping
from .makeStruc import makeBase, reCenter, randReorient

def atomsFillMesh(mesh, og, tol, radii, numMol):
    '''Fills mesh with single atoms'''
    filled = []

    if str(numMol).lower() == 'fill':
        numMol = 10000000000000

    # Create KD-tree for filledAtoms
    kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

    # Determine bounds of mesh
    bounds = mesh.bounds
    minBound, maxBound = bounds[0], bounds[1]

    # Determine spacing between molecules
    spacing = tol

    # Generate grid of points
    gridX, gridY, gridZ = np.mgrid[minBound[0]:maxBound[0]:spacing,
                                   minBound[1]:maxBound[1]:spacing,
                                   minBound[2]:maxBound[2]:spacing]

    # Check each point in the grid
    for x in np.nditer(gridX):
        for y in np.nditer(gridY):
            for z in np.nditer(gridZ):
                for atom in og:
                    # Construct atom data
                    atomData = [atom[0], x, y, z]
                    if len(atom) == 5:
                            atomData.append(atom[4])
                    point = [x, y, z]
                    if mesh.isInside(point) and \
                        (kdTree is None or not isOverlapAtomKDTree(atomData, kdTree, indexToAtom, radii, tol)):

                        filled.append(atomData)

                        # Rebuild KDTree with newly added atoms
                        kdTree, indexToAtom = buildKDTreeMapping(filled, radii)
    
    return filled

def moleculesFillMesh(mesh, og, tol, radii, numMol, baseStruc, 
                      randOrient):
    '''Fills mesh with molecules'''
    filled = []

    if str(numMol).lower() == 'fill':
        numMol = 10000000000000

    if baseStruc is not None:
        base = makeBase(baseStruc)
        filled.append(reCenter(base, mesh))

    # Create KD-tree for filledAtoms
    kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

    # Determine bounds of mesh
    bounds = mesh.bounds
    minBound, maxBound = bounds[0], bounds[1]

    # Determine spacing between molecules
    spacing = tol

    # Generate grid of points
    gridX, gridY, gridZ = np.mgrid[minBound[0]:maxBound[0]:spacing,
                                   minBound[1]:maxBound[1]:spacing,
                                   minBound[2]:maxBound[2]:spacing]
    
    # Check each point in the grid
    for i in range(gridX.shape[0]):
        for j in range(gridY.shape[1]):
            for k in range(gridZ.shape[2]):
                x = gridX[i, j, k]
                y = gridY[i, j, k]
                z = gridZ[i, j, k]

                # Check if entire molecule can be placed
                molValid = True
                newMol = []
                for atom in og:
                    # Construct atom data
                    atomType, relX, relY, relZ = atom[:4]
                    atomPoint = [x + relX, y + relY, z + relZ]

                    if not mesh.isInside(atomPoint):
                        molValid = False
                        break
                
                if molValid:
                    for atom in og:
                        atomType, relX, relY, relZ = atom[:4]
                        if len(atom) == 4:
                            atomData = (atomType, float(x + relX), float(y + relY), float(z + relZ))
                        if len(atom) == 5:
                            atomData = (atomType, float(x + relX), float(y + relY), float(z + relZ), atom[4])

                        newMol.append(atomData)
                        
                    if randOrient and len(newMol) == len(og):
                        newMol = randReorient(newMol)
                    if (kdTree is None or not isOverlapMoleculeKDTree(newMol, kdTree, indexToAtom, radii, tol)):
                        filled.append(newMol)

                        # Rebuild KDTree with newly added atoms
                        kdTree, indexToAtom = buildKDTreeMapping(filled, radii)
    return filled

def atomsRandMesh(mesh, og, tol, radii, numMol, maxAttempts):
    '''Randomly places atoms in a mesh'''
    filled = []
    attempts = 0

    # Create KD-tree for filledAtoms
    kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

    # Determine bounds of mesh
    bounds = mesh.bounds
    minBound, maxBound = bounds[0], bounds[1]

    while len(filled) < numMol and attempts <= maxAttempts:
        for atom in og:
            x = random.uniform(minBound[0], maxBound[0])
            y = random.uniform(minBound[1], maxBound[1])
            z = random.uniform(minBound[2], maxBound[2])
            atomType, xRel, yRel, zRel = atom[:4]
            atomPoint = [x + xRel, y + yRel, z + zRel]
            
            if mesh.isInside(atomPoint):
                if len(atom) == 4:
                    atomData = (atomType, atomPoint[0], atomPoint[1], atomPoint[2])

                elif len(atom) == 5:
                    atomData = (atomType, atomPoint[0], atomPoint[1], atomPoint[2], atom[4])

                if (kdTree is None or not isOverlapAtomKDTree(atomData, kdTree, indexToAtom, radii, tol)):
                    filled.append(atomData)

                    # Rebuild KDTree with newly added atoms
                    kdTree, indexToAtom = buildKDTreeMapping(filled, radii)
        attempts += 1
    return filled

def moleculesRandMesh(mesh, og, tol, radii, numMol, baseStruc,
                      randOrient, maxAttempts):
    '''Randomly places molecules in a mesh'''
    filled = []
    attempts = 0

    if baseStruc is not None:
        base = makeBase(baseStruc)
        filled.append(reCenter(base, mesh))

    # Create KD-tree for filledAtoms
    kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

    # Determine bounds of mesh
    bounds = mesh.bounds
    minBound, maxBound = bounds[0], bounds[1]

    while len(filled) < numMol and attempts <= maxAttempts:
        newMol = []

        x = random.uniform(minBound[0], maxBound[0])
        y = random.uniform(minBound[1], maxBound[1])
        z = random.uniform(minBound[2], maxBound[2])

        for atom in og:
            atomType, xRel, yRel, zRel = atom[:4]
            atomPoint = [x + xRel, y + yRel, z + zRel]

            if mesh.isInside(atomPoint):
                if len(atom) == 4:
                    atomData = (atomType, atomPoint[0], atomPoint[1], atomPoint[2])

                elif len(atom) == 5:
                    atomData = (atomType, atomPoint[0], atomPoint[1], atomPoint[2], atom[4])

                newMol.append(atomData)

        if randOrient and len(newMol) == len(og):
            newMol = randReorient(newMol)

        if (kdTree is None or not isOverlapMoleculeKDTree(newMol, kdTree, indexToAtom, radii, tol)):
            if len(newMol) == len(og):
                filled.append(newMol)

                # Rebuild KDTree with newly added atoms
                kdTree, indexToAtom = buildKDTreeMapping(filled, radii)

        attempts += 1
    return list(filled)


'''Module contains the main loop'''

import sys
import argparse
from drawMol import drawMolBox, drawMolSphere, drawMolMesh
from readWriteFiles import writeOutput, getInput, readStrucFile
from defaultParams import defaultParams

def parseCommandLine(dparams):
    '''Parses command line for parameters'''
    parser = argparse.ArgumentParser(description='Tile Fill')

    # Command line arguemnnts
    parser.add_argument('-i', '--inputFile', type=str, help="Path to input file")
    parser.add_argument('-struc', '--structureFile', type=str,
                        help='Path to structure file')
    parser.add_argument('-baseStruc', '--baseStrucFile', type=str,
                        help='Path to base structure file', default=dparams['baseStrucFile'])
    parser.add_argument('-s', '--shape', type=str,
                        help="Shape (Box or Cube)", default=dparams['shape'])
    parser.add_argument('-sl', '--sideLength', type=float,
                        help="Dimensions of a cube in Angstroms", default=dparams['sideLength'])
    parser.add_argument('-xl', '--Xlen', type=float,
                        help="X dimension of a box in Angstroms", default=dparams['Xlen'])
    parser.add_argument('-yl', '--Ylen', type=float,
                        help="Y dimension of a box in Angstroms", default=dparams['Ylen'])
    parser.add_argument('-zl', '--Zlen', type=float,
                        help="Z dimension of a box in Angstroms", default=dparams['Zlen'])
    parser.add_argument('-r', '--radius', type=float,
                        help="Radius of a sphere in Angstroms", default=dparams['radius'])
    parser.add_argument('-rand', '--randomizeOrient', type=bool,
                        help="Randomize the orientation or not", default=dparams['randomizeOrient'])
    parser.add_argument('-randFill', '--randFill', type=bool,
                        help="Randomize place molecules when placing a set number",
                        default=dparams['randFill'])
    parser.add_argument('-center', '--center', nargs='+', type=float,
                        help="X, Y, Z coordinates of the center of a sphere",
                        default=dparams['sphereCenter'])
    parser.add_argument('-baseCenter', '--baseStrucCenter', nargs='+', type=float,
                        help="X, Y, Z coordinates of the center of the base structure",
                        default=dparams['baseStrucCenter'])
    parser.add_argument('-n', '--numMolecules', type=int,
                        help="Number of molecules", default=dparams['numMolecules'])
    parser.add_argument('-t', '--tol', type=float,
                        help="Minimum distance between molecules", default=dparams['tol'])
    parser.add_argument('-rho', '--density', type=float,
                        help="Density in g/mL", default=dparams['density'])
    parser.add_argument('-a', '--maxAttempts', type=int,
                        help="Maximum iterations for finite sized systems",
                        default=dparams['maxAttempts'])
    parser.add_argument('-ar', '--atomRadius', type=str,
                        help="What radius type to be used for atoms",
                        default=dparams['atomRadius'])
    parser.add_argument('-out', '--outputFile', type=str,
                        help="Path for output file if desired")
    parser.add_argument('-mesh', '--mesh', type=str,
                        help="Path for mesh file if desired")
    parser.add_argument('-scale', '--meshScale', type=float,
                        help="Uniform scale of mesh", default=dparams['meshScale'])

    return parser.parse_args()

def main():
    '''The main loop'''
    # Init default params
    dparams = defaultParams()

    # Init arguement parser
    args = parseCommandLine(dparams)

    iparams = {}
    if args.inputFile:
        iparams = getInput(args.inputFile)
    else:
        for argName in vars(args):
            iparams[argName] = getattr(args, argName)

    for name in dparams:
        if name not in iparams:
            iparams[name] = dparams[name]

    if not iparams['structureFile']:
        print('Please provide an initial structure')
        sys.exit()

    print(iparams)

    # Get structure
    struc, unitCell = readStrucFile(iparams['structureFile'])
    print(struc)

    if unitCell:
        iparams['unitCell'] = [unitCell['a'], unitCell['b'], unitCell['c']]

    if iparams['baseStrucFile']:
        baseStruc, baseUnitCell = readStrucFile(iparams['baseStrucFile'])
    else:
        baseStruc = None

    if iparams['shape'].lower() == 'box' or iparams['shape'].lower() == 'cube':
        # Generate the new structure

        if unitCell:
            outStruc, strucType, cellParams = drawMolBox(struc, baseStruc, iparams)
        else:
            outStruc, strucType = drawMolBox(struc, baseStruc, iparams)

    elif iparams['shape'].lower() == 'sphere':
        # Generate the new structure

        if unitCell:
            outStruc, strucType, cellParams = drawMolSphere(struc, baseStruc, iparams)
        else:
            outStruc, strucType = drawMolSphere(struc, baseStruc, iparams)

    elif iparams['shape'].lower() == 'mesh' and iparams['mesh'] is not None:
        # Generate the new structure
        outStruc = []

#        if isinstance(iparams['mesh'], str):
#            iparams['mesh'] = [iparams['mesh']]
#
#        for i in iparams['mesh']:

        if unitCell:
            outStrucTemp, strucType, cellParams = drawMolMesh(struc, baseStruc, iparams)
            outStruc.append(outStrucTemp)
        else:
            outStrucTemp, strucType = drawMolMesh(struc, baseStruc, iparams)
            outStruc.append(outStrucTemp)

        
    if iparams['outputFile']:
        if unitCell:
            writeOutput(outStruc, iparams['outputFile'], strucType, cellParams)
        else:
            writeOutput(outStruc, iparams['outputFile'], strucType)

if __name__ == "__main__":
    main()



from setuptools import setup, find_packages
import versioneer

requirements = [
    'numpy',
    'pandas',
    'trimesh',
    'rtree',
    'scipy',
]

setup_requirements = []
test_requirements = requirements.append(['pytest'])

setup(
    python_requires = '>=3.7',
    packages = find_packages(include=['TurtleMol', 'TurtleMol.*']),
    install_requires=requirements,
    include_package_data=True,
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    zip_safe=False,
)


'''
This module checks if atoms within a molecule overlaps
or if two different molecules overlap
'''

import math
import numpy as np
import scipy.spatial

def isOverlapMolecule(newAtom, filledAtoms, radii, tol): # Depreciated!!!!
    '''Checks if two molecules overlap'''
    for molecule in filledAtoms:
        for atom1, atom2 in zip(newAtom, molecule):
            distance = math.sqrt(
                (atom1[1] - atom2[1])**2 +
                (atom1[2] - atom2[2])**2 +
                (atom1[3] - atom2[3])**2
            )

            if distance < (radii[atom1[0]] + radii[atom2[0]]) + tol:
                return True
    print('isOverlapMolecule is depreciated and will be removed from future versions!\n')
    return False

def isOverlapAtom(newAtom, filledAtoms, radii, tol): # Depreciated!!!!
    '''Checks if two atoms overlap'''
    for atom in filledAtoms:
        distance = math.sqrt(
            (newAtom[1] - atom[1])**2 +
            (newAtom[2] - atom[2])**2 +
            (newAtom[3] - atom[3])**2
        )

        if distance < (radii[newAtom[0]] + radii[atom[0]]) + tol:
            return True
        
    print('isOverlapMolecule is depreciated and will be removed from future versions!\n')
    return False

def isOverlapAtomKDTree(newAtom, kdTree, indexToAtom, radii, tol):
    '''Checks if two atoms overlap using a KDTree'''
    atomName, x, y, z = newAtom[:4]
    radius = radii[str.capitalize(atomName)]
    nearbyAtoms = kdTree.query_ball_point([x, y, z], radius + tol)

    for i in nearbyAtoms:
            molID, filledAtomName, filledAtomRadius = indexToAtom[i]
            otherAtom = kdTree.data[i]
            distance = np.linalg.norm(np.array([x, y, z]) - otherAtom)
            if distance < (radius + filledAtomRadius + tol):
                return True
    return False

def isOverlapMoleculeKDTree(newMol, kdTree, indexToAtom, radii, tol):
    '''Checks if any two atoms from a new molecule overlap using a KDTree'''
    for atom in newMol:
        atomName, x, y, z = atom[:4]
        radius = radii[str.capitalize(atomName)]
        nearbyAtoms = kdTree.query_ball_point([x, y, z], radius + tol)

        for i in nearbyAtoms:
            molID, filledAtomName, filledAtomRadius = indexToAtom[i]
            otherAtom = kdTree.data[i]
            distance = np.linalg.norm(np.array([x, y, z]) - otherAtom)
            if distance < (radius + filledAtomRadius + tol):
                return True
    return False

def buildKDTreeMapping(filledMolecules, radii):
    atomCoords = []
    indexToAtom = {}

    for molID, molecule in enumerate(filledMolecules):
        if isinstance(molecule, list):
            for atom in molecule:
                index = len(atomCoords)
                atomCoords.append(atom[1:4])
                indexToAtom[index] = (molID, atom[0], radii[atom[0]])
        else:
            index = len(atomCoords)
            atomCoords.append(molecule[1:4])
            indexToAtom[index] = (molID, molecule[0], radii[molecule[0]])

    kd_tree = scipy.spatial.cKDTree(atomCoords) if atomCoords else None
    return kd_tree, indexToAtom


'''
This module handles a 3d sphere to be filled
'''

import math

class Sphere3d:
    '''
    Defines the 3D Sphere object
    '''
    def __init__(self, xCoord, yCoord, zCoord, radius):
        self.xCoord = xCoord
        self.yCoord = yCoord
        self.zCoord = zCoord
        self.radius = radius

    def volume(self):
        '''Returns the volume of a Sphere'''
        return (4/3) * math.pi * self.radius**3

    def surfaceArea(self):
        '''Returns the surface area of a Sphere'''
        return 4 * math.pi * self.radius**2

    def containsPoints(self, xCoord, yCoord, zCoord, atomRadius):
        '''Returns if the distance between points is greater than the radius'''
        distance = math.sqrt((xCoord - self.xCoord)**2 +
                             (yCoord - self.yCoord)**2 +
                             (zCoord - self.zCoord)**2)
        return distance <= (self.radius + atomRadius)

    def findCenter(self):
        '''Find the center of the sphere'''
        return (self.xCoord, self.yCoord, self.zCoord)

    def origin(self):
        '''Returns the origin of the sphere'''
        return (self.xCoord, self.yCoord, self.zCoord)



'''
These functions are used to repeat a pattern of a given molecule or molecules. 
The goal is to repeat the pattern exactly the specified number of times
within the constraints of a the defined shape. 
A 'Fill' option was also added so that one could just define a space,
and fill said space with provided structure as many times as it can fit.
'''

import math
from .Box3D import Box3d, drawBox
from .Sphere3D import Sphere3d
from .mesh3D import mesh3D
from .shiftBox import atomsFillBox, atomsRandBox, \
                     moleculesFillBox, moleculesRandBox
from .shiftSphere import atomFillSphere, atomRandSphere, \
                        moleculeRandSphere, moleculeFillSphere
from .shiftMesh import atomsFillMesh, moleculesFillMesh, atomsRandMesh, moleculesRandMesh
from .setAtomProp import setAtomicRadius
from .makeStruc import makeBase, shiftPoints
from .shiftDensity import placeMols
from .shiftUnitCell import unitCellBox, unitCellMesh, unitCellSphere

def drawMolBox(struc, baseStruc, iparams):
    '''Utilized to place molecules in a box'''
    dims = drawBox(iparams)
    box = Box3d(0, 0, 0, dims)
    radii = setAtomicRadius(iparams['atomRadius'])
    numMol = iparams['numMolecules']
    tol = float(iparams['tol'])

    originalPoints = makeBase(struc)

    originalPoints = shiftPoints(originalPoints, box)

    if iparams['unitCell']:
        filled, strucType, cellParams = unitCellBox(box, dims, iparams['unitCell'], 
                                        originalPoints, radii)
        return filled, strucType, cellParams

    if iparams['density']:
        filled, strucType = placeMols(box, originalPoints, iparams['density'],
                                 tol, "box", radii, iparams['randomizeOrient'])
        return filled, strucType

    if not isinstance(numMol, int) and str(numMol).lower() != "fill":
        numMol = int(numMol)

    if iparams['randFill'] == 'False' or iparams['randFill'] is False:
        numXShifts = math.ceil(box.length / tol)
        numYShifts = math.ceil(box.height / tol)
        numZShifts = math.ceil(box.width / tol)

        # Check if structure is monatomic or molecule
        if len(originalPoints) == 1:
            filledAtom = atomsFillBox(numXShifts, numYShifts, numZShifts,
                                      tol, originalPoints, box, radii, numMol)
            return list(filledAtom), "atom"

        if len(originalPoints) > 1:
            filledAtom = moleculesFillBox(numXShifts, numYShifts, numZShifts,
                                          tol, originalPoints, box, radii, baseStruc,
                                          iparams['randomizeOrient'], numMol)
            return list(filledAtom), "molecule"

    elif iparams['randFill'] == 'True' or iparams['randFill'] is True:
        # Check if structure is monatomic or molecule
        if len(originalPoints) == 1:
            filledAtom = atomsRandBox(numMol, iparams['maxAttempts'],
                                          originalPoints, box, radii, tol)
            return list(filledAtom), "atom"

        if len(originalPoints) > 1:
            filledAtom = moleculesRandBox(numMol, iparams['maxAttempts'],
                                              originalPoints, box, radii, tol, baseStruc,
                                              iparams['randomizeOrient'])
            return list(filledAtom), "molecule"

    return "ERROR", "No atoms found"

def drawMolSphere(struc, baseStruc, iparams):
    '''Utilized to place molecules in a sphere'''
    sphere = Sphere3d(float(iparams['sphereCenter'][0]), float(iparams['sphereCenter'][1]),
                      float(iparams['sphereCenter'][2]), float(iparams['radius']))
    radii = setAtomicRadius(iparams['atomRadius'])
    numMol = iparams['numMolecules']
    tol = float(iparams['tol'])

    originalPoints = makeBase(struc)
    originalPoints = shiftPoints(originalPoints, sphere)

    if iparams['unitCell']:
        filled, strucType, cellParams = unitCellSphere(sphere, iparams['unitCell'], 
                                        originalPoints, radii)
        return filled, strucType, cellParams

    if iparams['density']:
        filled, strucType = placeMols(sphere, originalPoints, iparams['density'],
                                 tol, "sphere", radii, iparams['randomizeOrient'])
        return filled, strucType

    if not isinstance(numMol, int) and str(numMol).lower() != "fill":
        numMol = int(numMol)

    if iparams['randFill'] == 'False' or iparams['randFill'] is False:
        numShifts = math.ceil(2*sphere.radius / tol)

        # Check if structure is monatomic or molecule
        if len(originalPoints) == 1:
            filledAtom = atomFillSphere(numShifts, sphere,
                                        originalPoints, radii, tol,
                                        numMol)
            return list(filledAtom), "atom"

        if len(originalPoints) > 1:
            filledAtom = moleculeFillSphere(numShifts, sphere,
                                            originalPoints, radii, tol, baseStruc,
                                            iparams['randomizeOrient'], numMol)
            return list(filledAtom), "molecule"

    elif iparams['randFill'] == 'True' or iparams['randFill'] is True:
        # Check if structure is monatomic or molecule
        if len(originalPoints) == 1:
            filledAtom = atomRandSphere(numMol, iparams['maxAttempts'],
                                            originalPoints, sphere, radii, tol)
            return list(filledAtom), "atom"

        if len(originalPoints) > 1:
            filledAtom = moleculeRandSphere(numMol, iparams['maxAttempts'],
                                                originalPoints, sphere, radii, tol, baseStruc,
                                                iparams['randomizeOrient'])
            return list(filledAtom), "molecule"

    return "ERROR: No atoms found"

def drawMolMesh(struc, baseStruc, iparams):
    '''Places molecules in a mesh'''
    mesh = mesh3D(iparams['mesh'], iparams['meshScale'])
    radii = setAtomicRadius(iparams['atomRadius'])
    numMol = iparams['numMolecules']
    tol = float(iparams['tol'])

    originalPoints = makeBase(struc)

    if iparams['unitCell']:
        filled, strucType, cellParams = unitCellMesh(mesh, iparams['unitCell'], 
                                        originalPoints, radii)
        return filled, strucType, cellParams

    if iparams['density']:
        filled, strucType = placeMols(mesh, originalPoints, iparams['density'],
                                 tol, "mesh", radii, iparams['randomizeOrient'])
        return filled, strucType

    if not isinstance(numMol, int) and str(numMol).lower() != "fill":
        numMol = int(numMol)

    if iparams['randFill'] == 'False' or iparams['randFill'] is False:
        if len(originalPoints) == 1:
            filledAtom = atomsFillMesh(mesh, originalPoints, tol, radii, numMol)

            return list(filledAtom), 'atom'
        else:
            filledMol = moleculesFillMesh(mesh, originalPoints, tol,
                                          radii, numMol, baseStruc, iparams['randomizeOrient'])
            
            return list(filledMol), 'molecule'

    elif iparams['randFill'] == 'True' or iparams['randFill'] is True:
        if len(originalPoints) == 1:
            filledAtom = atomsRandMesh(mesh, originalPoints, tol, radii, numMol,
                                       iparams['maxAttempts'])
            return list(filledAtom), 'atom'
        else:
            filledMol = moleculesRandMesh(mesh, originalPoints, tol, radii, numMol,
                                          baseStruc, iparams['randomizeOrient'],
                                          iparams['maxAttempts'])
            return list(filledMol), 'molecule'


def defaultParams():
    '''
Keywords:
    shape = "Cube" or "Box" are the currently supported shapes
    sideLength = float(length of sides in a cube)
    Xlen = float(length of a box)
    Ylen = float(width of a box)
    Zlen = float(depth of a box)
    numMolecules = int(Number of molecules in box)
    tol = float(minimum distance between molecules)
    inputFile = Path to the input file if you are using one
    structureFile = Path to the structure file being used. 
        Required for usage. XYZ and PDB formats are currently supported
    outputFile = Path to output file if desired
    baseStrucFile = Path to file for a structure that is present, but
        not deuplicated
    baseStrucCenter = coordinates to place the geometric
        center of the base structure
    randomizeOrient = Boolean to randomize orientation of
        molecules
    randFill = When placing a set number of molecules, place them randomly
    desnity = total mass of molecules over the total volume
        of the system in g/mL
    atomRadius = Choose whether to use the AtomicRadius, CovalentRadius,
        or VanDerWaalsRadius

'''
    params = {
        'shape' : 'box',
        'sideLength' : 1.0,
        'Xlen' : 1.0,
        'Ylen' : 1.0,
        'Zlen' : 1.0,
        'sphereCenter' : [0, 0, 0],
        'radius' : 1.0,
        'numMolecules' : 1,
        'tol' : 1.0,
        'maxAttempts' : 10000,
        'baseStrucCenter' : [0, 0, 0],
        'baseStrucFile' : None,
        'randomizeOrient' : False,
        'randFill' : False,
        'density' : None,
        'atomRadius' : 'AtomicRadius',
        'mesh' : None,
        'meshScale' : 1.0,
        'unitCell' : None
    }

    return params



'''Sets up the structure files'''
import random
import numpy as np
from .Box3D import Box3d
from .setAtomProp import setAtomicMass

def makeBase(baseStruc):
    '''Convert dataframe to list of points'''
    atomBase = baseStruc['Atom'].values.tolist()
    xBase = baseStruc['X'].values.tolist()
    yBase = baseStruc['Y'].values.tolist()
    zBase = baseStruc['Z'].values.tolist()

    if 'Residue' not in baseStruc.columns:
        return [(atomBase[i], xBase[i], yBase[i], zBase[i]) for i in range(len(atomBase))]

    residue = baseStruc['Residue'].values.tolist()
    return [(atomBase[i], xBase[i], yBase[i], zBase[i], residue[i]) for i in range(len(atomBase))]

def calcCenter(coords):
    '''Calculate geometric center'''
    if not coords:
        return None

    numCoords = len(coords)
    centerX = sum(coord[1] for coord in coords) / numCoords
    centerY = sum(coord[2] for coord in coords) / numCoords
    centerZ = sum(coord[3] for coord in coords) / numCoords

    return (centerX, centerY, centerZ)

def reCenter(struc, shape):
    '''Set center structure coordiantes to center of shape'''
    currentCenter = calcCenter(struc)
    shapeCenter = shape.findCenter()
    displacement = (
        shapeCenter[0] - currentCenter[0],
        shapeCenter[1] - currentCenter[1],
        shapeCenter[2] - currentCenter[2]
    )
    newCoords = [(coord[0], coord[1] + displacement[0], coord[2] + displacement[1],
                coord[3] + displacement[2]) for coord in struc]

    return newCoords

def shiftPoints(points, shape):
    '''Sifts points to the bottom-left front corner of a 3D shape'''
    pointNames = [point[0] for point in points]
    coords = np.array([point[1:4] for point in points], dtype=float)

    if len(points[0]) == 5:
        pointRes = [point[4] for point in points]
    else:
        pointRes = None

    # Determine the bounds of the points
    minXPoints = np.min(coords[:, 0])
    minYPoints = np.min(coords[:, 1])
    minZPoints = np.min(coords[:, 2])

    # Calculate the translation required
    translation = np.array([
        shape.origin()[0] - minXPoints,
        shape.origin()[1] - minYPoints,
        shape.origin()[2] - minZPoints,
    ])

    # Shift the points
    shiftedCoords = coords + translation
    if pointRes:
        shiftedPoints = [[name] + coord.tolist() + [residue]
                         for name, coord, residue in
                         zip(pointNames, shiftedCoords, pointRes)]
    else:
        shiftedPoints = [[name] + coord.tolist()
                         for name, coord in zip(pointNames, shiftedCoords)]

    return shiftedPoints

def randReorient(mol):
    '''Randomly reorients a molecule around geometric center'''
    if len(mol) == 0:
        return mol

    atomNames = [atom[0] for atom in mol]
    atomInfo = [atom[4:] if len(atom) > 4 else [] for atom in mol]
    points = np.array([atom[1:4] for atom in mol], dtype=np.float64)
    centroid = points.mean(axis=0) # Calculate center of points
    points -= centroid # Translate the points to origin

    # Perform the rotaton
    # Generate random rotation angles in radians
    thetaX = random.uniform(0, 2*np.pi)
    thetaY = random.uniform(0, 2*np.pi)
    thetaZ = random.uniform(0, 2*np.pi)

    # Create the rotation matrix for the x-axis
    rx = np.array([[1, 0, 0],
                  [0, np.cos(thetaX), -np.sin(thetaX)],
                  [0, np.sin(thetaX), np.cos(thetaX)]],
                  dtype=np.float64)

    # Create rotation matrix for the y-axis
    ry = np.array([[np.cos(thetaY), 0, np.sin(thetaY)],
                  [0, 1, 0],
                  [-np.sin(thetaY), 0, np.cos(thetaY)]],
                  dtype=np.float64)

    # Create rotation matrix for the z-axis
    rz = np.array([[np.cos(thetaZ), -np.sin(thetaZ), 0],
                  [np.sin(thetaZ), np.cos(thetaZ), 0],
                  [0, 0, 1]],
                  dtype=np.float64)

    # Combined rotation matrix
    R = rz @ ry @ rx

    # Apply the rotation matrix to the points
    rotatedPoints = np.dot(points, R.T)

    # Translate the points back to position
    rotatedPoints += centroid

    rotatedAtoms = [tuple(name) + tuple(coord.tolist()) + tuple(info)
                    for name, coord, info in zip(atomNames, rotatedPoints, atomInfo)]

    rotatedAtoms = [tuple(atom) for atom in rotatedAtoms]

    return rotatedAtoms

def calcDensity(shape, mol):
    '''Calculates the density of a given structure'''
    vol = shape.volume() * 1e-24 # mL
    mass = 0
    atomicMass = setAtomicMass()  # g

    for atom in mol:
        mass += atomicMass[atom[0]]
    
    moles = mass # g

    return mass/vol # g/mL

def calcNumMol(shape, mol, density):
    '''
    Calulates the number of molecules needed in a box to match
    the defined density based off of the defined volume
    '''
    vol = shape.volume() # A^3
    rho = density # g/mL
    rho *= 1e-24 # g/A^3
    atomicMass = setAtomicMass() # g/mol
    molarMass = 0 
    for atom in mol:
        molarMass += atomicMass[atom[0]]
    
    numMol = (rho * 6.022e23 * vol)/molarMass # molecules

    return int(numMol)

def calcDistance(shape, mol, density, shapeType):
    '''
    Calculates the distance between molecules for the correct density
    '''
    if str(shapeType).lower() == "box":
        x, y, z = shape.length, shape.width, shape.height

        numMol = calcNumMol(shape, mol, density)

        # Calculate the molecules to place in each dimension
        pointsPerDimension = round(numMol ** (1/3))

        # Generate grid points
        XCount = np.linspace(0, x, pointsPerDimension, endpoint=False)
        YCount = np.linspace(0, y, pointsPerDimension, endpoint=False)
        ZCount = np.linspace(0, z, pointsPerDimension, endpoint=False)

        # Create a meshgrid and reshape it to get coordiantes
        xx, yy, zz = np.meshgrid(XCount, YCount, ZCount, indexing='ij')
        coords = np.vstack([xx.ravel(), yy.ravel(), zz.ravel()]).T

        return coords[:numMol].tolist()
    
    elif str(shapeType).lower() == "sphere":

        def insideSphere(x, y, z, radius):
            return x**2 + y**2 + z**2 <= radius**2

        r = shape.radius

        # Create cubic geometry to enclose the sphere
        cubeSide = 2*r

        box = Box3d(0, 0, 0, [cubeSide, cubeSide, cubeSide])

        numMol = calcNumMol(box, mol, density)

        # Calculate total points in the cube
        totalPoints = round(numMol ** (1/3))

        # Generate points within cube
        XCount = np.linspace(-r, r, int(totalPoints**(1/3)))
        YCount = np.linspace(-r, r, int(totalPoints**(1/3)))
        ZCount = np.linspace(-r, r, int(totalPoints**(1/3)))

        # Create meshgrid and reshape to get coordinates
        xx, yy, zz = np.meshgrid(XCount, YCount, ZCount, indexing='ij')
        coords = np.vstack([xx.ravel(), yy.ravel(), zz.ravel()]).T

        # Find points inside sphere
        insidePoints = [point for point in coords if insideSphere(*point, r)]
        
        return insidePoints[:numMol]
    
    elif str(shapeType).lower() == 'mesh':
        # Determine bounds of mesh
        bounds = shape.bounds
        minBound, maxBound = bounds[0], bounds[1]

        numMol = calcNumMol(shape, mol, density)

        # Calculate the molecules to place in each dimension
        pointsPerDimension = round(numMol ** (1/3))

        # Generate grid points
        XCount = np.linspace(minBound[0], maxBound[0], pointsPerDimension, endpoint=False)
        YCount = np.linspace(minBound[1], maxBound[1], pointsPerDimension, endpoint=False)
        ZCount = np.linspace(minBound[2], maxBound[2], pointsPerDimension, endpoint=False)

        # Create a meshgrid and reshape it to get coordiantes
        xx, yy, zz = np.meshgrid(XCount, YCount, ZCount, indexing='ij')
        coords = np.vstack([xx.ravel(), yy.ravel(), zz.ravel()]).T

        # Find points inside sphere
        insidePoints = [point for point in coords if shape.isInside(point)]

        return insidePoints[:numMol]